from infini.generator import TextGenerator
from infini.output import Output


def test_generator():
    generator = TextGenerator()
    generator.events = {
        "test.event1": "Event1 文本",
    }
    generator.match(Output("text", "test.event1"))
    assert generator.output(Output("text", "test.event1")) == "Event1 文本"


def test_generator_with_var():
    generator = TextGenerator()
    generator.events = {
        "test.event1": "Event1 文本: {{ var }}",
    }

    assert (
        generator.output(Output("text", "test.event1", variables={"var": "变量测试"}))
        == "Event1 文本: 变量测试"
    )


def test_generator_with_function():
    def add(a, b):
        return a + b

    generator = TextGenerator()
    generator.events = {
        "test.event": "{{ func(1, 2) }}",
    }

    assert (
        generator.output(Output("text", "test.event", variables={"func": add}))
        == "3"
    )



class Router:
    sign: str

    def __init__(self, sign: str) -> None:
        self.sign = sign

    def match(self, input: str) -> bool:
        return self.sign == input.strip()


class StartswithRouter(Router):
    def match(self, input: str) -> bool:
        return input.strip().startswith(self.sign)


class ContainsRouter(Router):
    def match(self, input: str) -> bool:
        return self.sign in input.strip()


class Endswith(Router):
    def match(self, input: str) -> bool:
        return input.strip().endswith(self.sign)



from infini.input import Input
from infini.output import Output
from infini.router import ContainsRouter, Router
from infini.typing import List, Dict, Any, Callable, RouterType
from functools import wraps


class Register:
    pre_interceptors: List[RouterType]
    handlers: List[RouterType]
    events: Dict[str, str]
    global_variables: Dict[str, str | Callable]
    interceptors: List[RouterType]

    def __init__(self) -> None:
        self.pre_interceptors = []
        self.handlers = []
        self.events = {}
        self.global_variables = {}
        self.interceptors = []

    def pre_interceptor(self, router: Router | str, priority: int = 0):
        def decorator(func):
            @wraps(func)
            def wrapper(input: Input) -> Input | Output:
                return func(input)

            self.pre_interceptors.append(
                {
                    "priority": priority,
                    "router": ContainsRouter(router)
                    if isinstance(router, str)
                    else router,
                    "handler": wrapper,
                }
            )
            return wrapper

        return decorator

    def handler(self, router: Router | str, priority: int = 0):
        def decorator(func):
            @wraps(func)
            def wrapper(input: Input) -> Output:
                return func(input)

            self.handlers.append(
                {
                    "priority": priority,
                    "router": ContainsRouter(router)
                    if isinstance(router, str)
                    else router,
                    "handler": wrapper,
                }
            )
            return wrapper

        return decorator

    def regist_textevent(self, name: str, text: str):
        self.events[name] = text

    def regist_variable(self, name: str, data: Any):
        self.global_variables[name] = data

    def dynamic_variable(self, name: str | None = None):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs) -> str:
                return func(*args, **kwargs)

            self.global_variables[name or func.__name__] = wrapper
            return wrapper

        return decorator

    def interceptor(self, router: Router | str, priority: int = 0):
        def decorator(func):
            @wraps(func)
            def wrapper(input: Input) -> Input | Output:
                return func(input)

            self.interceptors.append(
                {
                    "priority": priority,
                    "router": ContainsRouter(router)
                    if isinstance(router, str)
                    else router,
                    "handler": wrapper,
                }
            )
            return wrapper

        return decorator



from infini.loader import Loader
from infini.register import Register
from infini.typing import List
from pathlib import Path

import sys
import inspect


def require(name: str, paths: List | None = None) -> Register:
    caller_frame = inspect.stack()[1]
    caller_file = caller_frame[0].f_globals["__file__"]

    default_paths = [Path(caller_file).resolve().parent]
    paths = [
        str(path)
        for path in ((list(paths) + default_paths) if paths else default_paths)
    ]
    (sys.path.insert(0, path) for path in paths)

    with Loader() as loader:
        loader.load(name)
        sys.path = sys.path[len(paths) - 1 :]
        return loader.into_register()



from infini.output import Output
from infini.typing import Dict, Callable
from infini.exceptions import UnknownEvent
from jinja2 import Template


class TextGenerator:  # TODO 兼容多类型事件
    events: Dict[str, str]
    global_variables: Dict[str, str | Callable]

    def __init__(self) -> None:
        self.events = {}
        self.global_variables = {}

    def output(self, output: Output) -> str:
        assert output.type != "workflow", "Workflow 事件无法产出文本"
        variables = self.global_variables.copy()
        variables.update(output.variables)
        return self.match(output).render(variables)

    def match(self, output: Output) -> Template:
        if context := self.events.get(output.name):
            return Template(context)
        raise UnknownEvent(f"事件不存在: {output.name}")



from typing import (
    Dict as Dict,
    List as List,
    Any as Any,
    Generic as Generic,
    Callable as Callable,
    Literal as Literal,
    Sequence as Sequence,
    Generator as Generator,
    overload as overload,
    TypeVar,
    TypedDict,
    Union,
)
from types import ModuleType as ModuleType, GeneratorType as GeneratorType
from . import router, input, output

T = TypeVar("T")
Stream = Union["input.Input", "output.Output"]
OutputGenerator = Generator["output.Output", Any, None]


class RouterType(TypedDict):
    priority: int
    router: router.Router
    handler: Callable[["input.Input"], Union[Stream, OutputGenerator]]



from infini.input import Input
from infini.interceptor import Interceptor
from infini.generator import TextGenerator
from infini.handler import Handler
from infini.output import Output
from infini.typing import Any, Generator


class Core:
    pre_interceptor: Interceptor
    handler: Handler
    generator: TextGenerator
    interceptor: Interceptor

    def input(self, input: Input) -> Generator[str, Any, None]:
        for pre_intercepted_stream in self.pre_intercept(input):
            if isinstance(pre_intercepted_stream, Output):
                yield self.generate(pre_intercepted_stream)
                if pre_intercepted_stream.block or pre_intercepted_stream.is_empty():
                    return
            else:
                input = pre_intercepted_stream

        for handled_stream in self.handle(input):
            if handled_stream.is_empty():
                return
            outcome = self.generate(handled_stream)
            for stream in self.intercept(outcome):
                if isinstance(stream, Output):
                    yield self.generate(stream)
                    continue
                outcome = stream
            if handled_stream.block:
                return
            yield outcome

    def pre_intercept(self, input: Input) -> Generator[Output | Input, Any, None]:
        return self.pre_interceptor.input(input)

    def handle(self, input: Input):
        iterator = self.handler.input(input)
        for output in iterator:
            yield output

    def generate(self, output: Output) -> str:
        return self.generator.output(output)

    def intercept(self, output_text: str) -> Generator[Output | str, Any, None]:
        return self.interceptor.output(output_text)



from infini.typing import Dict, Any, Generic, T


class Input(Generic[T]):
    plain_data: T
    variables: Dict[str, Any]

    def __init__(self, plain_data: T, variables: Dict[str, Any] | None = None) -> None:
        self.plain_data = plain_data
        self.variables = variables or {}

    def get_session_id(self) -> str:
        if session_id := self.variables.get("session_id"):
            return session_id

        user_id = self.variables.get("user_id", "unknown")
        group_id = self.variables.get("group_id", "unknown")
        return f"session_{group_id}_{user_id}"

    def get_plain_text(self) -> str:
        return str(self.plain_data)



from infini.input import Input
from infini.output import Output
from infini.typing import List, Any, RouterType, Callable, Generator
from infini.queue import EventQueue


class Handler:
    handlers: List[RouterType]

    def input(self, input: Input):
        if (queue := self.match(input.get_plain_text())).is_empty():
            yield Output.empty()
            return
        while not queue.is_empty():
            if isinstance(stream := queue.pop()(input), Generator):
                for output in stream:
                    yield output
                    if output.block:
                        return
            else:
                yield stream
                if stream.block:
                    return

    def match(
        self, text: str
    ) -> EventQueue[Callable[[Input], Output | Generator[Output, Any, None]]]:
        queue = EventQueue()

        for handler in self.handlers:
            if handler["router"].match(text):
                queue.push(handler["priority"], handler["handler"])

        return queue



from infini.input import Input
from infini.output import Output
from infini.typing import List, Any, RouterType, Callable, Generator
from infini.queue import EventQueue


class Interceptor:
    interceptors: List[RouterType]

    def input(self, input: Input) -> Generator[Output | Input, Any, None]:
        queue = self.match(input.get_plain_text())
        while not queue.is_empty():
            if isinstance(stream := queue.pop()(input), Generator):
                for outcome in stream:
                    if isinstance(outcome, Input):
                        input = outcome
                        break
                    yield outcome
                    if outcome.block:
                        return
            else:
                if stream is None:
                    yield Output.empty()
                    return
                if isinstance(stream, Output):
                    yield stream
                    if stream.block:
                        return
                    continue
                input = stream
        yield input

    def output(
        self, output_text: str
    ) -> Generator[Output | str, Any, None]:
        input = Input(output_text)
        queue = self.match(input.get_plain_text())
        while not queue.is_empty():
            if isinstance(stream := queue.pop()(input), Generator):
                for outcome in stream:
                    if isinstance(outcome, Input):
                        input = outcome
                        break
                    yield outcome
                    if outcome.block:
                        return
            else:
                if stream is None:
                    yield Output.empty()
                    return
                if isinstance(stream, Output):
                    yield stream
                    if stream.block:
                        return
                    continue
                input = stream
        yield input.get_plain_text()

    def match(
        self, text: str
    ) -> EventQueue[
        Callable[[Input], Input | Output | Generator[Input | Output, Any, None]]
    ]:
        queue = EventQueue()

        for interceptor in self.interceptors:
            if interceptor["router"].match(text):
                queue.push(interceptor["priority"], interceptor["handler"])

        return queue
